import numpy as np
import scipy


def interpolate(val1, time1, val2, time2, time):
    interp = scipy.interpolate.interp1d([time1, time2], [val1, val2])

    return interp(time)


def rolling_window_test(a, window, window_type, time=None, distance=None):
    return _rolling_window(a, window, window_type, time, distance)


# window_type= None-> #features; 'time'-> time based; 'distance'-> distance based
def _rolling_window(a, window, window_type, time=None, distance=None):
    if window_type is None:
        shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
        strides = a.strides + (a.strides[-1],)
        for row in np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides):
            yield row
    elif window_type == 'time':  # TODO: optimize
        if time is None:
            raise ValueError(f"time cannot be None when window_type={window_type}")
        for i in range(len(time - 1)):
            total_time = 0
            values = []

            for j, (corr, succ) in enumerate(zip(time[i:-1], time[i + 1:])):
                delta = succ - corr
                if total_time + delta > window:
                    interpolated_value = interpolate(a[i + j], corr, a[i + j + 1], succ, window)
                    values.append(interpolated_value)
                    break
                else:
                    total_time += delta
                    values.append(a[i + j])
            yield values
    elif window_type == 'distance':  # TODO: optimize
        if distance is None:
            raise ValueError(f"distance cannot be None when window_type={window_type}")
        for i in range(len(distance - 1)):
            total_distance = 0
            values = []

            for j, corr, succ in enumerate(zip(distance[i:-1], distance[i + 1:])):
                delta = succ - corr
                if total_distance + delta > window:
                    interpolated_value = interpolate(a[i + j], corr, a[i + j + 1], succ, window)
                    values.append(interpolated_value)
                    break
                else:
                    total_distance += delta
                    values.append(a[i + j])
            yield values
    else:
        raise ValueError(f"window_type={window_type} unsupported. window_type must be in [None, 'time', 'distance']")


def sum(features: np.ndarray, window, window_type=None, time=None, distance=None):
    if window is None:
        window = len(features)

    # np.sum(_rolling_window(features, window, window_type))

    return np.sum(
        np.fromiter(_rolling_window(features, window, window_type, time, distance), dtype=np.dtype((float, window))),
        axis=1)


def max(features: np.ndarray, window, window_type=None, time=None, distance=None):
    if window is None:
        window = len(features)

    res = []

    for array in _rolling_window(features, window, window_type, time, distance):
        res.append(np.max(array))

    return np.array(res)


def min(features: np.ndarray, window, window_type=None, time=None, distance=None):
    if window is None:
        window = len(features)

    res = []

    for array in _rolling_window(features, window, window_type, time, distance):
        res.append(np.min(array))

    return np.array(res)


def std(features: np.ndarray, window, window_type=None, time=None, distance=None):
    if window is None:
        window = len(features)

    res = []

    for array in _rolling_window(features, window, window_type, time, distance):
        res.append(np.std(array))

    return np.array(res)


def cov(features: np.ndarray, window, window_type=None, time=None, distance=None):
    if window is None:
        window = len(features)

    res = []

    for array in _rolling_window(features, window, window_type, time, distance):
        res.append(np.cov(array))

    return np.array(res)


def var(features: np.ndarray, window, window_type=None, time=None, distance=None):
    if window is None:
        window = len(features)

    res = []

    for array in _rolling_window(features, window, window_type, time, distance):
        res.append(np.var(array))

    return np.array(res)


def rate_upper(features: np.ndarray, threshold, window, distance=None, time=None, window_type=None):
    if window is None:
        window = len(features)

    returnValue = []

    """count = 0
    for i in range(len(features)):

        if features[i] > threshold:
            count += 1
        if i - window + 1 > 0 and features[i - window] > threshold:
            count -= 1
        if i - window + 1 >= 0:
            returnValue[i - window + 1] = count"""

    for i, window_value in enumerate(_rolling_window(features, window, window_type, time, distance)):
        returnValue.append(0)
        for el in window_value:
            if el > threshold:
                returnValue[i] += 1

    returnValue = np.array(returnValue)

    if distance is not None:
        returnValue /= distance

    return returnValue


def rate_below(features: np.ndarray, threshold, window, distance=None, time=None, window_type=None):
    return rate_upper(features=features, threshold=threshold * -1, window=window, window_type=window_type,
                      distance=distance, time=time)



import numpy as np

import BasicFeatures
import AggregateFeatures


def compute_all(lat: np.ndarray, lon: np.ndarray, time: np.ndarray, window, threshold,
                window_type=[None, 'time', 'distance']):
    basic_features = []

    basic_features.append(BasicFeatures.speed(lat=lat, lon=lon, time=time))
    basic_features.append(BasicFeatures.acceleration(lat=lat, lon=lon, time=time))
    basic_features.append(BasicFeatures.acceleration2(lat=lat, lon=lon, time=time))
    basic_features.append(BasicFeatures.distance(lat=lat, lon=lon))
    basic_features.append(BasicFeatures.direction(lat=lat, lon=lon))

    if type(threshold) != list:
        threshold = [threshold for _ in basic_features]

    aggregated_features = None
    aggregated_features_time = None
    aggregated_features_distance = None

    index = window_type.index(None) if None in window_type else None
    if index is not None:
        aggregated_features = []
        for features, threshold_value in zip(basic_features, threshold):
            aggregated_features.append(AggregateFeatures.sum(features=features, window=window[index]))
            aggregated_features.append(AggregateFeatures.max(features=features, window=window[index]))
            aggregated_features.append(AggregateFeatures.min(features=features, window=window[index]))
            aggregated_features.append(AggregateFeatures.std(features=features, window=window[index]))
            aggregated_features.append(AggregateFeatures.cov(features=features, window=window[index]))
            aggregated_features.append(AggregateFeatures.var(features=features, window=window[index]))
            aggregated_features.append(AggregateFeatures.rate_upper(features=features, threshold=threshold_value,
                                                                    window=window[index], distance=basic_features[3]))
            aggregated_features.append(AggregateFeatures.rate_below(features=features, threshold=threshold_value,
                                                                    window=window[index], distance=basic_features[3]))
    index = window_type.index('distance') if None in window_type else None
    if index is not None:
        aggregated_features_distance = []
        distance = basic_features[3]
        for features, threshold_value in zip(basic_features, threshold):
            aggregated_features_distance.append(
                AggregateFeatures.sum(features=features, window=window[index], window_type='distance', distance=distance))
            aggregated_features_distance.append(
                AggregateFeatures.max(features=features, window=window[index], window_type='distance', distance=distance))
            aggregated_features_distance.append(
                AggregateFeatures.min(features=features, window=window[index], window_type='distance', distance=distance))
            aggregated_features_distance.append(
                AggregateFeatures.std(features=features, window=window[index], window_type='distance', distance=distance))
            aggregated_features_distance.append(
                AggregateFeatures.cov(features=features, window=window[index], window_type='distance', distance=distance))
            aggregated_features_distance.append(
                AggregateFeatures.var(features=features, window=window[index], window_type='distance', distance=distance))
            aggregated_features_distance.append(
                AggregateFeatures.rate_upper(features=features, threshold=threshold_value,
                                             window=window[index], distance=distance, window_type='distance'))
            aggregated_features_distance.append(
                AggregateFeatures.rate_below(features=features, threshold=threshold_value,
                                             window=window[index], distance=distance, window_type='distance'))

    index = window_type.index('time') if None in window_type else None
    if index is not None:
        aggregated_features_time = []
        for features, threshold_value in zip(basic_features, threshold):
            aggregated_features_time.append(
                AggregateFeatures.sum(features=features, window=window[index], window_type='time', time=time))

            aggregated_features_time.append(
                AggregateFeatures.max(features=features, window=window[index], window_type='time', time=time))

            aggregated_features_time.append(
                AggregateFeatures.min(features=features, window=window[index], window_type='time', time=time))

            aggregated_features_time.append(
                AggregateFeatures.std(features=features, window=window[index], window_type='time', time=time))

            aggregated_features_time.append(
                AggregateFeatures.cov(features=features, window=window[index], window_type='time', time=time))

            aggregated_features_time.append(
                AggregateFeatures.var(features=features, window=window[index], window_type='time', time=time))

            aggregated_features_time.append(
                AggregateFeatures.rate_upper(features=features, threshold=threshold_value, window=window[index],
                                             time=time, window_type='time'))
            aggregated_features_time.append(
                AggregateFeatures.rate_below(features=features, threshold=threshold_value, window=window[index],
                                             time=time, window_type='time'))

    return basic_features, aggregated_features, aggregated_features_time, aggregated_features_distance



import numpy as np
import geopy.distance as gp

def speed(lat: np.ndarray, lon: np.ndarray, time: np.ndarray):
    d_time = time[1:] - time[:-1]
    dist_lat = lat[1:] - lat[:-1]
    dist_lon = lon[1:] - lon[:-1]
    dist = (dist_lat ** 2 + dist_lon ** 2) ** .5

    return np.append([.0], dist / d_time)


def acceleration(lat: np.ndarray, lon: np.ndarray, time: np.ndarray):
    d_time = time[1:] - time[:-1]
    dist_lat = lat[1:] - lat[:-1]
    dist_lon = lon[1:] - lon[:-1]
    dist = (dist_lat ** 2 + dist_lon ** 2) ** .5

    return np.append([.0], dist / (d_time ** 2))


def acceleration2(lat: np.ndarray, lon: np.ndarray, time: np.ndarray):
    d_time = time[1:] - time[:-1]
    dist_lat = lat[1:] - lat[:-1]
    dist_lon = lon[1:] - lon[:-1]
    dist = (dist_lat ** 2 + dist_lon ** 2) ** .5
    acc = np.append([.0], dist / (d_time ** 2))

    return np.append([.0], acc[1:] - acc[:-1])


def direction(lat: np.ndarray, lon: np.ndarray):
    dist_lat = lat[1:] - lat[:-1]
    dist_lon = lon[1:] - lon[:-1]

    return np.append([.0], np.arctan2(dist_lat, dist_lon))


def distance(lat: np.ndarray, lon: np.ndarray, ellipsoid="WGS-84"):
    returnValue = np.zeros(len(lat) - 1)
    for i in range(len(lat) - 1):
        returnValue[i] = gp.geodesic((lat[i], lon[i]), (lat[i + 1], lon[i + 1]), ellipsoid=ellipsoid).meters

    return np.append([.0], returnValue)
